#include "oxprob.h"
#include "sv96a.ox"

static decl train_start  = 444; 	// correspond to Jul 1963
static decl train_length = 953;		// correspond to Dec 2005
static decl test_length  = 35;

static decl pi = 3.1415926536;

decl d_post;
decl k_post;

// Class for posterior of d	======================================================
class postd
{
	decl psi, Ainv, Qs, k, T, q;
	post_d(const d);
	set_other(const new_psi, const new_Ainv, const new_Qs, const new_k, const new_T, const new_q);
};

postd::post_d(const d)
{
	decl Cinv_d = zeros(q,q);
	decl prod_d = new array[T];
		 prod_d[0:T-1] = zeros(q,q);
	  
	decl Q_sqrt, Qu, Qd, Qv;
	for(decl t=0; t<T; ++t)
	{
		decsvd(Qs[t][][], &Qu, &Qd, &Qv);
		Q_sqrt = Qu * diag(Qd.^(d/2)) * Qv';
 
		prod_d[t][][] = Q_sqrt * invert(Qs[t+1][][]) * Q_sqrt;
		Cinv_d += prod_d[t][][];
	}
	 
	return psi*d - 0.5*k*trace(Ainv*Cinv_d);
}

postd::set_other(const new_psi, const new_Ainv, const new_Qs, const new_k, const new_T, const new_q)
{
	psi  = new_psi;
	Ainv = new_Ainv;
	Qs   = new_Qs;
	k    = new_k;
	T    = new_T;
	q    = new_q;
}

post_d(const d)
{
	return d_post -> post_d(d);
}
//================================================================================

// Class for posterior of k ======================================================
class postk
{
	decl lambda0, A, Cinv_k, sum_log_Q, T, q;
	post_k(const k);
	set_other_k(const new_lambda0, const new_A, const new_Cinv_k, const new_sum_log_Q, const new_T, const new_q); 
};

postk::post_k(const k)
{
	return	-lambda0*k + 0.5*T*k*q*log(k/2) - 0.5*T*k*log(determinant(A))
			- T*sumr(log(gammafact((k+1-range(1,q))/2))) + (k/2)*sum_log_Q - 0.5*k*trace(invert(A)*Cinv_k); 
}

postk::set_other_k(const new_lambda0, const new_A, const new_Cinv_k, const new_sum_log_Q, const new_T, const new_q)
{
	lambda0 = new_lambda0;
	A         = new_A;
	Cinv_k    = new_Cinv_k;
	sum_log_Q = new_sum_log_Q;
	T         = new_T;
	q         = new_q;
}

post_k(const k)
{
	return k_post -> post_k(k);
}
//================================================================================

// Wishart generator =============================================================
rwish(const df, const Sc)
{
	decl p = sizer(Sc);
	decl B = 0 * Sc;
 
	if(p>1)
	{
		decl i,j;
		for(j=0; j<p; ++j)
		{
			B[j][j] = sqrt( rangamma(1, 1, (df-j)/2, 1/2) );   
		}

		for(i=1; i<p; ++i)
		{
			for(j=0; j<i; ++j)
			{
				B[i][j] = rann(1,1);
			}
		}   
	}
 
	decl A = choleski(Sc);
 
	return A * B * B' * A';
}
//================================================================================

// Set initial values for log-vols ===============================================
H(const phi, const mu, const seta, const T)
{
	decl i;
	decl lag     = 1;
	decl q       = sizer(phi);
	decl eta     = new matrix[q][T];
	decl sig_eta = diag(seta);

	for(i=0; i<T; i++)
	{
		eta[][i] = choleski(sig_eta) * rann(q,1);
	}
	
	decl h = eta[][1] ~ new matrix[q][T];

	for(i=lag; i<(T+lag); i++)
	{
		h[][i] = mu + phi .* h[][i-1] + eta[][i-1];
	}

	decl vol = h[][1:(T+lag-1)];
	
	return(vol);
}
//================================================================================

// Set initial values for Qt and Sigmat ==========================================
Initials(const T, const A, const k, const d)
{
	decl q = sizer(A);
	// Storages	========================
	decl initial = new array[2][T+1];
	initial[0][0:T] = zeros(q,q);
	initial[1][0:T] = zeros(q,q);
	 //==================================

	decl Q = unit(q);
	decl Qu, Qd, Qv;
		 decsvd(Q, &Qu, &Qd, &Qv);

	decl Q_neg_sqrt = Qu*diag(Qd.^(-d/2))*Qv';
	decl S = (1/k) * Q_neg_sqrt * A * Q_neg_sqrt;

	initial[0][0] = Q;
	initial[1][0] = Q;
	
	decl Su, Sd, Sv, Qinv, Qinv_star_can, Sigmat;
	for(decl i=1; i<=T; ++i)
	{
		Qinv = rwish(k, S);   
		Q = invert(Qinv);

		Qinv_star_can = diag(diagonal(Q).^-.5);
		Sigmat = Qinv_star_can * Q * Qinv_star_can;
		
		decsvd(Q, &Qu, &Qd, &Qv);
		Q_neg_sqrt = Qu*diag(Qd.^(-d/2))*Qv';
		S = (1/k) * Q_neg_sqrt * A * Q_neg_sqrt;

		initial[0][i] = Q;
		initial[1][i] = Sigmat;
	}
	
	return initial;
}
//================================================================================

// Set multivariate density function =============================================
densnorm(const y, const mu, const sig2)
{
	decl p = sizec(mu);
	
	return ( (2*pi)^(-p/2) )*(determinant(sig2))^(-1/2) * exp( (-1/2) * (y-mu)' * invert(sig2) * (y-mu) );
}
//================================================================================


main()
{
	// Most general settings
	decl time = timer(), j, t, i;
	decl n_samples = 30000;
	decl burn      = 10000;
	decl mod       = 5000;
	decl fuller_correct	= 0.00001;
	
	decl pred_length = test_length+1;
	
	//This part is to calculate the true variance and VaR
	decl Y_Orig     = .01 * loadmat("C:/Research/F-F data/Sec42_Portfolio_Returns.csv");  // raw_Y is T by p
	decl p          = sizec(Y_Orig);
	
	decl test_start = train_length - train_start + 1;
	decl test_end   = test_start + test_length;

	decl hold      = (1/sizec(Y_Orig)) * ones(sizec(Y_Orig),1);

	// Model Comparison and Assessment ===========================================
	decl discrep = zeros(pred_length, 1);
	decl loglik  = zeros(pred_length, 1);
	decl eq_wght = zeros(pred_length, 1);	
	// ===========================================================================	

	// To store the results ======================================================
	decl keep_A_det_all = zeros(n_samples-burn);
	decl keep_d_all     = zeros(n_samples-burn);
	decl keep_k_all     = zeros(n_samples-burn);
	decl elapse = 0;
	decl elapse_5k = zeros(n_samples/mod + 1, 1);	
	// ===========================================================================

	
	for(decl hh = 0; hh < pred_length; hh++)
	{
	
		//Store Draws ===========================================
		decl keep_A_det = zeros(n_samples,1);
		decl keep_d     = zeros(n_samples,1);
		decl keep_k     = zeros(n_samples,1);
		decl t_5k       = 0;		
		//=======================================================

		
		// Model Comparison and Assessment =======================================
		decl discrep_val = zeros(n_samples-burn,1);
		decl loglik_val  = zeros(n_samples-burn,1);
		decl eq_wght_val = zeros(n_samples-burn,1);			
		// =======================================================================

		decl end_Y = train_length + hh;
		decl Y_0   = Y_Orig[end_Y+1][]';

		// AM_Observations ========================================================
		decl Y = Y_Orig[train_start:end_Y][];  // raw_Y is T by p
		decl T = sizer(Y);

		decl acc = zeros(T,1);

		// Observed Factors
		decl F    = .01 * loadmat("C:/Research/F-F data/Sec42_FF_Factors.csv")[train_start:end_Y][1:3];	// F is T by q
		
		decl F1 = F'[0][],			// F1 is 1 by T
			 F2 = F'[1][],			// F2 is 1 by T
			 F3 = F'[2][];			// F2 is 1 by T
			 
		decl q  = sizec(F);			// number of factors 
	
		//============================================================================

		// For SVOL parameters
		decl ytrans1 = 1.2704 + log((F1.^2) + fuller_correct);
		decl ytrans2 = 1.2704 + log((F2.^2) + fuller_correct);
		decl ytrans3 = 1.2704 + log((F3.^2) + fuller_correct);
	
		decl n = columns(ytrans1);
	
    	decl mV,mF,mN,mSignal;     		//KF stuff and log-volatility
    
		decl para_keep = zeros(q,3);	// records MCMC results
    	decl beta_result = new array[q];
		 	 beta_result[0:(q-1)] = zeros(3, (n_samples-burn));

		 
		// For covariance prediction
		decl VarYT      = zeros(p*p, 1);	
	
		// Set hyperparameters and initial values =====================================
		decl B_0, L_0;			// Borrow the strength of MCMCpack to get B0 and Omega0
		B_0 = zeros(p,q);
		L_0 = unit(q);
		 
		decl invL_0	= invert(L_0);
		decl post_mean_B;
		decl post_var_B;		 

		// initial value for B
		decl B  = rann(p,q);
	
		// Omega is diagonal with each element following a scaled inverted chi-square
		// initial value for Omega	
		decl Osig = 0.2 * ones(p,1);

		// hyperparameters of the Omega diag elements
		decl nu = 10;
		decl s  = 0.01;

	
		// SVOL equations
		decl mu1 = 0,
			 mu2 = 0,
			 mu3 = 0;
		 
		decl mu  = (0~0~0)';
	
		decl phi = zeros(3,1);

		decl seta = zeros(3,1);

		decl gamma = 20, delta = 1.5;               
		decl df = 5;
		decl S_0 = df * 0.01;

		decl mIndex1, mIndex2, mIndex3;
	
		decl means1 = zeros(1,n),
			 means2 = zeros(1,n),
			 means3 = zeros(1,n);
	
		decl mScales1 = 4.93 * ones(1,n),
			 mScales2 = 4.93 * ones(1,n),
			 mScales3 = 4.93 * ones(1,n);	;
	
		decl w1, var1, sig1, pi1;
		decl w2, var2, sig2, pi2;
		decl w3, var3, sig3, pi3;
	
		SV_Mixture_Setup(&w1,&var1,&sig1,&pi1);
		SV_Mixture_Setup(&w2,&var2,&sig2,&pi2);
		SV_Mixture_Setup(&w3,&var3,&sig3,&pi3);
	
	    decl iInitial = 150;

		decl gilks1 = GW_SV(F1, fuller_correct, 10, iInitial);
		decl gilks2 = GW_SV(F2, fuller_correct, 10, iInitial);
		decl gilks3 = GW_SV(F3, fuller_correct, 10, iInitial);
	
		beta_result[0][0][0:iInitial-10-1] = log(gilks1[0][]./(1 - gilks1[0][]));	
		beta_result[0][1][0:iInitial-10-1] = 2*log(gilks1[1][]);
	
		beta_result[1][0][0:iInitial-10-1] = log(gilks2[0][]./(1 - gilks2[0][]));	
		beta_result[1][1][0:iInitial-10-1] = 2*log(gilks2[1][]);

		beta_result[2][0][0:iInitial-10-1] = log(gilks3[0][]./(1 - gilks3[0][]));	
		beta_result[2][1][0:iInitial-10-1] = 2*log(gilks2[1][]);

	
		decl mVaug, mVt=zeros(1,n);
	    decl nill = zeros(1,n);

	    decl vpfinish1 = meanr(beta_result[0][0:1][0:iInitial-10-1]),
	    	 vpfinish2 = meanr(beta_result[1][0:1][0:iInitial-10-1]),
			 vpfinish3 = meanr(beta_result[2][0:1][0:iInitial-10-1]);
	
		decl mHess1 = 2 .* variance(beta_result[0][0:1][0:iInitial-10-1]'),
			 mHess2 = 2 .* variance(beta_result[1][0:1][0:iInitial-10-1]'),
			 mHess3 = 2 .* variance(beta_result[2][0:1][0:iInitial-10-1]');
	
		decl seta1 = exp(vpfinish1[1][0]),
			 seta2 = exp(vpfinish2[1][0]),
			 seta3 = exp(vpfinish3[1][0]);

		decl phi1  = exp(vpfinish1[0][0]) / (1 + exp(vpfinish1[0][0])),
			 phi2  = exp(vpfinish2[0][0]) / (1 + exp(vpfinish2[0][0])),
			 phi3  = exp(vpfinish3[0][0]) / (1 + exp(vpfinish3[0][0]));
	

		seta  =	(seta1 ~ seta2 ~ seta3)';
		phi   = (phi1 ~ phi2 ~ phi3)';
	
	    para_keep = zeros(3,1) ~ phi ~ sqrt(seta);
	
		decl mHessinv1 = invertsym(mHess1),
			 mHessinv2 = invertsym(mHess2),
			 mHessinv3 = invertsym(mHess3);

	    decl like_est1, like_est2, like_est3;
		 
		decl vpstore1 = zeros(2,2),
			 vpstore2 = zeros(2,2),
			 vpstore3 = zeros(2,2);
		 
	    decl likes1 = -1.0e16 + zeros(1,2),
			 likes2 = -1.0e16 + zeros(1,2),
			 likes3 = -1.0e16 + zeros(1,2);
	
		decl Metro_lag1 = zeros(1,2),
			 Metro_lag2 = zeros(1,2),
			 Metro_lag3 = zeros(1,2); 	

		decl vp1 = zeros(2,1),
			 vp2 = zeros(2,1),
			 vp3 = zeros(2,1);


		decl phi1_init = 0.97, phi2_init = 0.97, phi3_init = 0.97;
		decl seta1_init = 0.2, seta2_init = 0.2, seta3_init = 0.2;
	
		vp1[0] = log(phi1_init/(1 - phi1_init));
		vp1[1] = log(seta1_init^2);

		vp2[0] = log(phi2_init/(1 - phi2_init));
		vp2[1] = log(seta2_init^2);

		vp3[0] = log(phi3_init/(1 - phi3_init));
		vp3[1] = log(seta3_init^2);
	
		vpstore1[0][] = vp1';
		vpstore1[1][] = vp1';

		vpstore2[0][] = vp2';
		vpstore2[1][] = vp2';

		vpstore3[0][] = vp3';
		vpstore3[1][] = vp3';
	
	
		decl mResult1, mResult2, mResult3, dLik1, dLik2, dLik3, dVar1, dVar2, dVar3;	// initialise SSFpack stuff
		decl mPhi1   = zeros(3,2),
			 mPhi2   = zeros(3,2),
			 mPhi3   = zeros(3,2);
		 
	    decl mOmega1 = diag(<0.01, 0, 4.93>),
			 mOmega2 = diag(<0.01, 0, 4.93>),
			 mOmega3 = diag(<0.01, 0, 4.93>);
		
		decl mDelta  = zeros(3,1);
	
		decl mSigma1 = zeros(3,2),
			 mSigma2 = zeros(3,2),
			 mSigma3 = zeros(3,2);

		decl mJ_Phi   = <-1,-1;-1,-1;-1,-1>;
		decl mJ_Omega = <-1,-1,-1;-1,-1,-1;-1,-1,0>;
		decl mJ_Delta = <-1;-1;-1>;	

		mSigma1[0][0] = seta1 / (1 - phi1^2) + 1.0e6;
		mSigma1[0][1] = 1.0e6;
		mSigma1[1][0] = 1.0e6;
		mSigma1[1][1] = 1.0e6;

	    mSigma2[0][0] = seta2 / (1 - phi2^2) + 1.0e6;
		mSigma2[0][1] = 1.0e6;
		mSigma2[1][0] = 1.0e6;
		mSigma2[1][1] = 1.0e6;

		mSigma3[0][0] = seta3 / (1 - phi3^2) + 1.0e6;
		mSigma3[0][1] = 1.0e6;
		mSigma3[1][0] = 1.0e6;
		mSigma3[1][1] = 1.0e6;
	
		mPhi1[0][0]   = phi1;
		mPhi1[0][1]   = 1 - phi1;
		mPhi1[1][1]   = 1;
		mPhi1[2][0]   = 1;

		mPhi2[0][0]   = phi2;
		mPhi2[0][1]   = 1 - phi2;
		mPhi2[1][1]   = 1;
		mPhi2[2][0]   = 1;
	
		mPhi3[0][0]   = phi3;
		mPhi3[0][1]   = 1 - phi3;
		mPhi3[1][1]   = 1;
		mPhi3[2][0]   = 1;

		mOmega1[0][0] = seta1;
		mOmega2[0][0] = seta2;
		mOmega3[0][0] = seta3;
	
		decl mean_lin1 = ytrans1 - means1,
			 mean_lin2 = ytrans2 - means2,
			 mean_lin3 = ytrans3 - means3; 
	
		// initializing log-volatilities with ran standard normal
		decl mSignal1 = H(phi, mu, seta, T)[0][]';
		decl mSignal2 = H(phi, mu, seta, T)[1][]'; 
		decl mSignal3 = H(phi, mu, seta, T)[2][]';
	
		// Correlation Level				   
		decl gamma_0 = q;
		decl C0_inv  = gamma_0 * unit(q);
		decl lambda0 = .02;
	
		decl A  	 = unit(q);
		decl k  	 = zeros(1,1); k = 20;
		decl d  	 = zeros(1,1); d = 0.75;  //d = 0.75
		decl Qt 	 = Initials(T, A, k, d)[0][];	
		decl Sigmat  = Initials(T, A, k, d)[1][];
		decl epsilon = new matrix[q][T];
	
		decl V = new array[T];		// initialize V_t
			 V[0:(T-1)] = zeros(q,q);
		 
		decl z = new matrix[q][T];	// z is q by T

		for(t=0; t<T; ++t)
		{
			V[t]   = diag( exp(mSignal1[t]) ~ exp(mSignal2[t]) ~ exp(mSignal3[t]) );
			z[][t] = invert(choleski(V[t])) * F[t][]';
		}
	
		//========================================================

		// General settings for d and k
		decl ninit_d    = 3;
		decl ninit_k    = 3; 	// This is for ARS. When log-concavity is assured, we required fewer support points.
	//	decl ninit_k    = 5; 	// This is for ARMS, as we do not know log-concavity, more initial support points are required
								// See Martino, Read, and Luengo (2012)
		decl dometrop_d = 1;
		decl dometrop_k = 0;
		decl nsamp      = 1;

		//For d
		decl mdl    = zeros(1,1); mdl[0][0]   = -0.99;
		decl mdr    = zeros(1,1); mdr[0][0]   =  0.99;
		decl dprev  = zeros(1,1); dprev[0][0] =  0.75;
		decl samp_d = zeros(1,nsamp); 	
		decl ansd   = zeros(1,1);	
			 d_post = new postd();	                    

		//For k							
		decl mkl    = zeros(1,1); mkl[0][0]   = q + 0.1;
		decl mkr    = zeros(1,1); mkr[0][0]   = 95;
		decl kprev  = zeros(1,1); kprev[0][0] = 20;
		decl samp_k = zeros(1,nsamp);
 		decl ansk   = zeros(1,1);  
			 k_post = new postk();
	
		//Start the sampling!
		for(j=0; j<n_samples; ++j)
		{						
			// B, Omega
			post_var_B  = invert(F'*F + invL_0);
			post_mean_B = (Y'*F + B_0*invL_0) * post_var_B;

			B = vec(post_mean_B) + choleski(diag(Osig) ** post_var_B) * rann(p*q,1);		
			B = reshape(B, q, p)';
			
			for(decl l=0; l<p; ++l)
			{		
				Osig[l] = 1/rangamma(1, 1, (T + nu)/2, 1/2) * ( nu * s + ( Y[][l] - F * B[l][]' )'*( Y[][l] - F * B[l][]' ) );
			}
					
			decl Ome = diag(Osig);
		
			// SVOL
			// Update SVOL parameters
			vpstore1[1][]  = (vpfinish1 + (choleski(mHess1) * rann(2,1)))';
			vpstore2[1][]  = (vpfinish2 + (choleski(mHess2) * rann(2,1)))';
	        vpstore3[1][]  = (vpfinish3 + (choleski(mHess3) * rann(2,1)))';
		
	        mean_lin1 = ytrans1 - means1;
			mean_lin2 = ytrans2 - means2;
			mean_lin3 = ytrans3 - means3;

			for(decl jj=0; jj<2; jj++)
			{
				seta1 = exp(vpstore1[jj][1]);
		    	seta2 = exp(vpstore2[jj][1]);
		    	seta3 = exp(vpstore3[jj][1]);			
			
				phi1  = exp(vpstore1[jj][0]) / (1 + exp(vpstore1[jj][0]));
				phi2  = exp(vpstore2[jj][0]) / (1 + exp(vpstore2[jj][0]));
				phi3  = exp(vpstore3[jj][0]) / (1 + exp(vpstore3[jj][0]));
			
				mSigma1[0][0] = seta1 / (1- phi1^2) + 1.0e6;
				mSigma2[0][0] = seta2 / (1- phi2^2) + 1.0e6;
				mSigma3[0][0] = seta3 / (1- phi3^2) + 1.0e6;
			
				mPhi1[0][0]   = phi1;
				mPhi1[0][1]   = 1 - phi1;
			
				mPhi2[0][0]   = phi2;
				mPhi2[0][1]   = 1 - phi2;

				mPhi3[0][0]   = phi3;
				mPhi3[0][1]   = 1 - phi3;
			
				mOmega1[0][0] = seta1;
				mOmega2[0][0] = seta2;
				mOmega3[0][0] = seta3;			
			
				SsfLik(&dLik1, &dVar1, mean_lin1, mPhi1, mOmega1, mSigma1, mDelta, mJ_Phi, mJ_Omega ,mJ_Delta, mScales1);
				SsfLik(&dLik2, &dVar2, mean_lin2, mPhi2, mOmega2, mSigma2, mDelta, mJ_Phi, mJ_Omega ,mJ_Delta, mScales2);
				SsfLik(&dLik3, &dVar3, mean_lin3, mPhi3, mOmega3, mSigma3, mDelta, mJ_Phi, mJ_Omega ,mJ_Delta, mScales3);
			
				likes1[0][jj] = dLik1;
				likes2[0][jj] = dLik2;
				likes3[0][jj] = dLik3;
			

				like_est1    = likes1[0][jj] + (gamma - 1) * log(1 + phi1) + (delta - 1) * log(1 - phi1)
							   - (S_0 / (seta1 * 2)) - (((df/2) + 1)*log(seta1));        

				like_est2    = likes2[0][jj] + (gamma - 1) * log(1 + phi2) + (delta - 1) * log(1 - phi2)
							   - (S_0 / (seta2 * 2)) - (((df/2) + 1)*log(seta2));

				like_est3    = likes3[0][jj] + (gamma - 1) * log(1 + phi3) + (delta - 1) * log(1 - phi3)
							   - (S_0 / (seta3 * 2)) - (((df/2) + 1)*log(seta3));
						   
				Metro_lag1[0][jj] = like_est1 - (MultiDen(vpstore1[jj][]', vpfinish1, mHessinv1) 
			                         - log(seta1) - log(phi1*(1 - phi1)));

				Metro_lag2[0][jj] = like_est2 - (MultiDen(vpstore2[jj][]', vpfinish2, mHessinv2) 
			                         - log(seta2) - log(phi2*(1 - phi2)));

				Metro_lag3[0][jj] = like_est3 - (MultiDen(vpstore3[jj][]', vpfinish3, mHessinv3) 
			                         - log(seta3) - log(phi3*(1 - phi3)));								 
			}


			if (log(ranu(1,1)) < (Metro_lag1[0][1] - Metro_lag1[0][0]) && phi1 < 0.995)
			{
				para_keep[0][0] = mu1;		
				para_keep[0][1] = phi1;
				para_keep[0][2] = sqrt(seta1);
				vpstore1[0][]   = vpstore1[1][];
				likes1[0][0]    = likes1[0][1];
			}
			else
			{
				seta1 = para_keep[0][2]^2;
				phi1  = para_keep[0][1];
			}

		
			if (log(ranu(1,1)) < (Metro_lag2[0][1] - Metro_lag2[0][0]) && phi2 < 0.995)
			{
				para_keep[1][0] = mu2;		
				para_keep[1][1] = phi2;
				para_keep[1][2] = sqrt(seta2);
				vpstore2[0][]   = vpstore2[1][];
				likes2[0][0]    = likes2[0][1];
			}
			else
			{
				seta2 = para_keep[1][2]^2;
				phi2  = para_keep[1][1];
			}
		
			if (log(ranu(1,1)) < (Metro_lag3[0][1] - Metro_lag3[0][0]) && phi3 < 0.995)
			{
				para_keep[2][0] = mu3;		
				para_keep[2][1] = phi3;
				para_keep[2][2] = sqrt(seta3);
				vpstore3[0][]   = vpstore3[1][];
				likes3[0][0]    = likes3[0][1];
			}
			else
			{
				seta3 = para_keep[2][2]^2;
				phi3  = para_keep[2][1];
			}
		
			mSigma1[0][0] = seta1 / (1 - phi1^2) + 1.0e6;
			mSigma2[0][0] = seta2 / (1 - phi2^2) + 1.0e6;
			mSigma3[0][0] = seta3 / (1 - phi3^2) + 1.0e6;
		
	    	mPhi1[0][0]   = phi1;
			mPhi1[0][1]   = 1 - phi1;
			mPhi2[0][0]   = phi2;
			mPhi2[0][1]   = 1 - phi2;
			mPhi3[0][0]   = phi3;
			mPhi3[0][1]   = 1 - phi3;		
		
			mOmega1[0][0] = seta1;     
			mOmega2[0][0] = seta2;
			mOmega3[0][0] = seta3;

			// Update the log-vol
			mResult1 = SsfCondDens(ST_SIM, mean_lin1, mPhi1, mOmega1, mSigma1, mDelta, mJ_Phi, mJ_Omega, mJ_Delta, mScales1);
			mResult2 = SsfCondDens(ST_SIM, mean_lin2, mPhi2, mOmega2, mSigma2, mDelta, mJ_Phi, mJ_Omega, mJ_Delta, mScales2);
			mResult3 = SsfCondDens(ST_SIM, mean_lin3, mPhi3, mOmega3, mSigma3, mDelta, mJ_Phi, mJ_Omega, mJ_Delta, mScales3);
		
			mSignal1 = mResult1[2][];
			mSignal2 = mResult2[2][];
			mSignal3 = mResult3[2][];
		
			// To get the residuals	epsilon_t = z_t
			for(t=0; t<T; ++t)
			{
				V[t]   = diag( exp(mSignal1[t]) ~ exp(mSignal2[t]) ~ exp(mSignal3[t]) );
				z[][t] = invert(choleski(V[t])) * F[t][]';	
			}	

			mu1 = mResult1[1][1];
			mu2 = mResult2[1][1];
			mu3 = mResult3[1][1];
		
	        MixtureUpdate(ytrans1, mSignal1, w1, var1, sig1, pi1, &mIndex1);		    
	        Mean_var(mIndex1, w1, var1, &means1, &mScales1);

			MixtureUpdate(ytrans2, mSignal2, w2, var2, sig2, pi2, &mIndex2);		    
	        Mean_var(mIndex2, w2, var2, &means2, &mScales2);
		
	        MixtureUpdate(ytrans3, mSignal3, w3, var3, sig3, pi3, &mIndex3);		    
	        Mean_var(mIndex3, w3, var3, &means3, &mScales3);
		
			// Sample the correlation level			
			// Update Qt, t = 1, 2,..., T-1
			decl Qu, Qd, Qv, Qu_can, Qd_can, Qv_can;  
			decl Q, Q_neg_sqrt, S, S_hat;
			decl canQt, canQtinv, Q_negsqrt_can, S_can, Qinv_star_can, P_can;
			decl Q_cur, Q_negsqrt_cur, Qu_cur, Qd_cur, Qv_cur;
			decl Qinv_star_cur, S_cur, P_cur;
			decl f_Qinv, f_Qinv_c;
				
			for(t=1; t<T; ++t)
			{
				Q = Qt[t-1];
				decsvd(Q, &Qu, &Qd, &Qv);
				Q_neg_sqrt = Qu * diag(Qd.^(-d/2)) * Qv';
				S = (1/k) * Q_neg_sqrt * A * Q_neg_sqrt;		
				S_hat = S - (1/(1 + z[][t-1]'*S*z[][t-1])) * (S*z[][t-1]*z[][t-1]'*S);	//Sherman-Morrison formula
//				S_hat = invert(invert(S) + z[][t-1] * z[][t-1]');						//Usual Inversion

			
				//Use Self Writtern Wishart DGP and svd
				canQtinv = rwish(k+1, S_hat);                  //Use Wishart generator
				canQt    = invert(canQtinv);
			
				//Original definition  
				decsvd(canQt, &Qu_can, &Qd_can, &Qv_can);		   //For candidate.
				Q_negsqrt_can = Qu_can * diag(Qd_can.^(-d/2)) * Qv_can';
				S_can         = (1/k) * Q_negsqrt_can * A * Q_negsqrt_can; 	
				Qinv_star_can = diag(diagonal(canQt).^-.5);
				P_can         = Qinv_star_can * canQt * Qinv_star_can;

				f_Qinv        = sqrt(determinant(invert(P_can))) * ( determinant(canQtinv)^((-1-k*d)/2) ) *		 
								exp((-0.5) * trace(invert(S_can) * invert(Qt[t+1])) -
								0.5*trace( (invert(P_can)-canQtinv) * z[][t-1] * z[][t-1]'));

				Q_cur         = Qt[t];				 
				decsvd(Q_cur, &Qu_cur, &Qd_cur, &Qv_cur);
				Q_negsqrt_cur = Qu_cur * diag(Qd_cur.^(-d/2)) * Qv_cur';
				S_cur         = (1/k) * Q_negsqrt_cur * A * Q_negsqrt_cur;
				Qinv_star_cur = diag(diagonal(Q_cur).^-.5);			   
				P_cur         = Qinv_star_cur * Q_cur * Qinv_star_cur;
	  
				f_Qinv_c      = sqrt(determinant(invert(P_cur))) * ( determinant(invert(Q_cur))^((-1-k*d)/2) ) *	  
								exp((-0.5) * trace(invert(S_cur) * invert(Qt[t+1])) -
								0.5*trace((invert(P_cur)-invert(Q_cur)) * z[][t-1] * z[][t-1]'));	

				decl Rt_cur, Rt_can;				
			
				decl R = f_Qinv / f_Qinv_c;

				if(R > ranu(1,1))
				{
					Qt[t] = canQt;

					if(j >= burn)
					{
						acc[t-1] = acc[t-1] + 1;
					}					
				}
			}							

				
			//Update QT
			decl Qu_T, Qd_T, Qv_T;
			decl ST_can, ST_hat;
			decl canQT, canQT_inv, Q_negsqrt_can_T, Qinv_star_can_T, PT_can;
			decl QT_cur, Q_negsqrt_cur_T, Qinv_star_cur_T, PT_cur, f_QinvT, f_QinvT_c;
   
			ST_hat = S_cur - (1/(1 + z[][T-1]'*S_cur*z[][T-1])) * (S_cur*z[][T-1]*z[][T-1]'*S_cur);
//			ST_hat = invert(invert(S_cur) + z[][T-1]*z[][T-1]');
			canQT_inv  = rwish(k+1, ST_hat); 
			canQT      = invert(canQT_inv);

			decsvd(canQT, &Qu_T, &Qd_T, &Qv_T);
			Q_negsqrt_can_T = Qu_T * diag(Qd_T.^(-d/2)) * Qv_T';
			ST_can          = (1/k) * Q_negsqrt_can_T * A * Q_negsqrt_can_T;
			Qinv_star_can_T = diag(diagonal(canQT).^-.5);
			PT_can          = Qinv_star_can_T * canQT * Qinv_star_can_T;

			f_QinvT         = (determinant(PT_can)^-.5) * (determinant(canQT_inv)^-.5) * 
							   exp( (-0.5)*trace( (invert(PT_can)-canQT_inv) * z[][T-1] * z[][T-1]' ) );
						  
			//current
			QT_cur          = Qt[T];
			Qinv_star_cur_T = diag(diagonal(QT_cur).^-.5);
			PT_cur          = Qinv_star_cur_T * QT_cur * Qinv_star_cur_T;
		
			f_QinvT_c       = (determinant(PT_cur)^-.5)* (determinant(invert(QT_cur))^-.5) *
							   exp( (-0.5)*trace( (invert(PT_cur)-invert(QT_cur)) * z[][T-1] * z[][T-1]' ) );
						   
			// To calculate the Var_t+1						   
			decl RT_cur, RT_can, f_cur, f_can, y_cur, y_can, p_y_cur, p_y_can;
			decl y_LC_cur, y_LC_can, p_y_LC_cur, p_y_LC_can;
		
			decl Q_T_plus_1_cur = invert(rwish(k, S_cur));
			decl Q_T_plus_1_can = invert(rwish(k, S_can));

			decl P_T_plus_1_cur = diag(diagonal(Q_T_plus_1_cur).^-.5) * Q_T_plus_1_cur * diag(diagonal(Q_T_plus_1_cur).^-.5);
			decl P_T_plus_1_can = diag(diagonal(Q_T_plus_1_can).^-.5) * Q_T_plus_1_can * diag(diagonal(Q_T_plus_1_can).^-.5);

			decl V_sq = diag(exp( (mu1 + phi1*(mSignal1[0][T-1]-mu1))/2 + sqrt(seta1/4)*rann(1,1)) ~
							 exp( (mu2 + phi2*(mSignal2[0][T-1]-mu2))/2 + sqrt(seta2/4)*rann(1,1)) ~
							 exp( (mu3 + phi3*(mSignal3[0][T-1]-mu3))/2 + sqrt(seta3/4)*rann(1,1)) );
						 
			if(j >= burn)
			{
				RT_cur = V_sq * P_T_plus_1_cur * V_sq;
			
				f_cur   = choleski(RT_cur) * rann(q,1);

				y_cur   = B * f_cur + (Ome.^.5) * rann(p,1);
				discrep_val[(j-burn)] = (y_cur - Y_0)' * (y_cur - Y_0);
			
				y_LC_cur   = hold' * B * f_cur + sqrt(hold' * Ome * hold) * rann(1,1);
				p_y_LC_cur = densnorm( (hold' * Y_0), (hold' * B * f_cur), (hold' * Ome * hold));
				eq_wght_val[(j-burn)] = log(p_y_LC_cur);
			
				p_y_cur = densnorm(Y_0, B*f_cur, Ome);
				loglik_val[(j-burn)]  = log(p_y_cur);
			
				while(p_y_cur < exp(-50))
				{
					f_cur   = choleski(RT_cur) * rann(q,1);

					y_cur   = B * f_cur + (Ome.^.5) * rann(p,1);
					discrep_val[(j-burn)] = (y_cur - Y_0)' * (y_cur - Y_0);
			
					y_LC_cur   = hold' * B * f_cur + sqrt(hold' * Ome * hold) * rann(1,1);
					p_y_LC_cur = densnorm( (hold' * Y_0), (hold' * B * f_cur), (hold' * Ome * hold));
					eq_wght_val[(j-burn)] = log(p_y_LC_cur);
			
					p_y_cur = densnorm(Y_0, B*f_cur, Ome);
					loglik_val[(j-burn)]  = log(p_y_cur);				
				}
			}
	  
			decl RT = f_QinvT / f_QinvT_c;
                       
			if(RT > ranu(1,1))
			{                                 
				Qt[T] = canQT;
		
				if(j >= burn)
				{
					acc[T-1] = acc[T-1] + 1;
					RT_can = V_sq * P_T_plus_1_can * V_sq;			
 
					f_can = choleski(RT_can) * rann(q,1);				
					y_can = B * f_can + (Ome.^.5) * rann(p,1);
					
					discrep_val[(j-burn)] = (y_can - Y_0)' * (y_can - Y_0);

					y_LC_can   = hold' * B * f_can + sqrt(hold' * Ome * hold) * rann(1,1);
					p_y_LC_can = densnorm( (hold' * Y_0), (hold' * B * f_can), (hold' * Ome * hold));
					eq_wght_val[(j-burn)] = log(p_y_LC_can);
				
					p_y_can = densnorm(Y_0, B*f_can, Ome);
					loglik_val[(j-burn)] = log(p_y_can);
				
					while(p_y_can < exp(-50))
					{
						f_can = choleski(RT_can) * rann(q,1);				
						y_can = B * f_can + (Ome.^.5) * rann(p,1);
					
						discrep_val[(j-burn)] = (y_can - Y_0)' * (y_can - Y_0);

						y_LC_can   = hold' * B * f_can + sqrt(hold' * Ome * hold) * rann(1,1);
						p_y_LC_can = densnorm( (hold' * Y_0), (hold' * B * f_can), (hold' * Ome * hold));
						eq_wght_val[(j-burn)] = log(p_y_LC_can);
				
						p_y_can = densnorm(Y_0, B*f_can, Ome);
						loglik_val[(j-burn)] = log(p_y_can);
					}
				}
			}

								
			//Update A ==================================================================================   
			decl A_inv, Au, Ad, Av, Q_sqrt, Cinv_sum = zeros(q,q);
			decl prod_A = new array[T];
			 	 prod_A[0:T-1] = zeros(q,q);
		
			for(decl t=0; t<T; ++t)
			{
				decsvd(Qt[t], &Au, &Ad, &Av);
				Q_sqrt    = Au * diag(Ad.^(d/2)) * Av';
				prod_A[t] = Q_sqrt * invert(Qt[t+1]) * Q_sqrt;
				Cinv_sum += prod_A[t];
			}
	 
			decl Cinv = k * Cinv_sum;

			A_inv = rwish((k*T + gamma_0), invert(C0_inv + Cinv));
			A     = invert(A_inv);
			keep_A_det[j] = determinant(A);			
			//===========================================================================================

			//Update d ==================================================================================
			decl sum_log_det = 0;			 // to compute psi
			for(decl t=1; t<=T; ++t)
			{
				sum_log_det += log(determinant(invert(Qt[t-1])));
			}
		
			decl psi = (-k/2) * sum_log_det;

			d_post -> set_other(psi, A_inv, Qt, k, T, q);

			ansd = Arms_simple(post_d, ninit_d, mdl, mdr, dometrop_d, dprev, samp_d);

			dprev[0][0] = samp_d[0][0];
			d = samp_d[0][0];
			keep_d[j]   = d;			
			//===========================================================================================

			//Update k ==================================================================================  
			decl sum_log_Qt = 0;
			decl Cinvk = zeros(q,q);
			decl prod_k = new array[T];
			 	 prod_k[0:T-1] = zeros(q,q);
	  
			decl Q_sqrt_k, Qu_k, Qd_k, Qv_k;
			for(decl t=0; t<T; ++t)
			{
				decsvd(Qt[t], &Qu_k, &Qd_k, &Qv_k);
				Q_sqrt_k = Qu_k * diag(Qd_k.^(d/2)) * Qv_k';
 
				prod_k[t]   = Q_sqrt_k * invert(Qt[t+1]) * Q_sqrt_k;
				sum_log_Qt += log(determinant(prod_k[t]));
				Cinvk      += prod_k[t][][];
			}

			k_post -> set_other_k(lambda0, A, Cinvk, sum_log_Qt, T, q);

			ansk = Arms_simple(post_k, ninit_k, mkl, mkr, dometrop_k, kprev, samp_k);
	
			kprev[0][0] = samp_k[0][0];
			k = samp_k[0][0];
			keep_k[j]   = k;

			if(fmod(j+1, mod) == 0)
			{
				t_5k = t_5k | (timer()-time)/100;
			}			
		}

		// Overall results ========================================================
		keep_A_det_all = keep_A_det_all ~ keep_A_det[burn:(n_samples-1)];
		keep_d_all     = keep_d_all ~ keep_d[burn:(n_samples-1)];
		keep_k_all     = keep_k_all ~ keep_k[burn:(n_samples-1)];

		elapse_5k      = elapse_5k ~ t_5k;		
		elapse = elapse | (timer()-time)/100;
		// ========================================================================

		
		// ========================================================================	
 		print("=====  pred  ", 36 - test_length + hh, "  =================", "\n");
//		print("elapse: ", elapse, "\n");	
	
		discrep[hh] = (1/(n_samples-burn)) * sumc(discrep_val);
		loglik[hh]  = (1/(n_samples-burn)) * sumc(loglik_val);
		eq_wght[hh] = (1/(n_samples-burn)) * sumc(eq_wght_val);
		
		print("discrep = ", discrep[hh], "\n");
		print("loglik = ",  loglik[hh], "\n");	
		print("loglik LC = ", eq_wght[hh], "\n");
	
		print("\n");
		print("acc rate=", (1/(n_samples-burn)) * acc, "\n");		
	}

//	savemat("C:/Research/Empirical/ARMS_NoSMW_A.csv",       keep_A_det_all);
//	savemat("C:/Research/Empirical/ARMS_NoSMW_d.csv",       keep_d_all);
//	savemat("C:/Research/Empirical/ARMS_NoSMW_k.csv",       keep_k_all);
//	savemat("C:/Research/Empirical/ARMS_NoSMW_time.csv",    elapse);	
//	savemat("C:/Research/Empirical/ARMS_NoSMW_time_5k.csv", elapse_5k);	
	
	print("Discrepency = ",   discrep, "\n");
	print("Pred Log-like = ", loglik,  "\n");
	print("Pred Log-like LC = ", eq_wght, "\n");

	print("Mean Discrepency = ",   meanc(discrep), "\n");
	print("Mean Pred Log-like = ", meanc(loglik), "\n");
	print("Mean Pred Log-like LC = ", meanc(eq_wght), "\n");
	
	print("time elapsed = ", (timer()-time)/100, "\n");
	print("This is AM Three-Fact, ten portfolios, ARS + SMW", "\n");
//	print("This is AM Three-Fact, ten portfolios, ARMS + NoSMW", "\n");	
	return 0;
}
